import argparse

from .__init__ import CACHE_PATH, PORT


def kill():
    import socket

    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.sendto(b'2Kill', ('localhost', PORT))


def clear():
    import os

    try:
        os.remove(CACHE_PATH)
    except FileNotFoundError:
        pass


def main():
    parser = argparse.ArgumentParser(
        prog='shellserver'
    )

    parser.add_argument(
        'command',
        choices=('kill', 'clear'),
        help='"kill" to kill the server, "clear" to clear the cache.'
    )

    args = parser.parse_args()

    if args.command == 'kill':
        kill()
    elif args.command == 'clear':
        clear()



from __future__ import annotations

"""
Module for functions with high level of abstraction
and/or complexity.
"""

import time
from fnmatch import fnmatchcase
# fnmatch.fnmatch would unnecessarily replace path.sep

from . import low
from . import medium

os = low.os  # os module
sys = medium.sys

n = '\n\n'
if '--git-verbose' not in sys.argv:
    # sending to stderr would break tests
    # and I have already messed with sys.stdout in server.py
    def print(*args, **kwargs):
        pass


class DirEntryWrapper:
    # os.DirEntry is a final class, can't be subclassed
    __slots__ = 'entry', 'name', 'path'

    def __init__(self, entry: os.DirEntry):
        self.entry = entry
        self.name = entry.name.lower()
        self.path = entry.path.lower()

    def stat(self):
        try:
            return self.entry.stat()
        except PermissionError:
            return self.entry.stat(follow_symlinks=False)

    def is_file(self):
        try:
            return self.entry.is_file()
        except PermissionError:
            return self.entry.is_file(follow_symlinks=False)

    def is_dir(self):
        try:
            return self.entry.is_dir()
        except PermissionError:
            return self.entry.is_dir(follow_symlinks=False)

    def is_symlink(self):
        return self.entry.is_symlink()


class FallbackError(Exception):
    pass


class High(medium.Medium):

    __slots__ = ('git_dir', 'branch', 'fixed', 'ignored', 'packs_list'
                 'untracked', 'staged', 'modified', 'deleted')

    def __init__(self) -> None:
        self.untracked = 0
        self.staged = 0
        self.modified = 0
        self.deleted = 0

    def init(self, git_dir, branch):
        self.git_dir = git_dir.lower()
        self.branch = branch

    def status(self):
        """
        Do a inline version of 'git status' without any use of git.
        return: str | None: String with the status, None if there's
                            nothing to report.
        """

        print('GIT_DIR:', self.git_dir)
        print('Branch', self.branch, end=n)

        try:
            self.set_index_tracked()
        except medium.IndexTooBigError:
            print("Index too big, Fallback.")
            raise FallbackError

        self.set_split_ignored()
        self.set_packs()
        print('Ignored patterns:')
        for pat in self.ignored:
            print(pat)
        for pat in self.fixed:
            print(pat)
        print()

        print('List of packfiles:')
        for pat in self.packs_list:
            print(pat)
        print()

        self.get_full_status(recurse_path=self.git_dir)

        status_string = self.get_status_string(
            (self.untracked, self.staged, self.modified, self.deleted)
        )

        self.untracked = 0
        self.staged = 0
        self.modified = 0
        self.deleted = 0

        return status_string

    def get_full_status(
        self,
        recurse_path,
        tree_hash=None
    ) -> tuple[int, int, int, int]:
        """
        Carries all the workload of getting a status.
        param `recurse_path`: For recursive use.
        param `tree_hash`: For recursive use.
        return: tuple: tuple of four integer numbers representing the number of
                        untracked, staged, modified and deleted files.
        """

        tree_items_list = []
        # will happen only in very first call
        if tree_hash is None:
            last_cmmt = self.get_last_commit_hash()
            if last_cmmt is not None:
                res = self.get_tree_items(last_cmmt)
                if res is not None:
                    tree_items_list = res
                    print('Found:', *tree_items_list, sep='\n', end=n)
        # all recursive calls will fall here
        else:
            tree_items_list = self.get_tree_items_from_hash(tree_hash)

        try:
            directory = os.scandir(recurse_path)
        # if symlink (a file) points to a dir
        except (PermissionError, NotADirectoryError):
            return

        # index will have only the full path of blobs
        # trees have everything
        found_in_tree = 0
        for file in directory:

            file = DirEntryWrapper(file)

            relpath = file.path.removeprefix(
                self.git_dir + '\\'
            ).replace('\\', '/', -1)

            # 100755(exe) is file
            if file.is_file() and relpath in self.index_tracked:
                found_in_tree += self.handle_tracked_file(
                    file, tree_items_list, relpath
                )

            elif any(fnmatchcase(file.name, pattern)
                     for pattern in self.ignored):
                pass
            elif any(fnmatchcase(
                    relpath + '/' if pattern.endswith('/') else relpath,
                    pattern) for pattern in self.fixed):
                pass

            elif file.is_file():  # and relpath not in index
                print('Untracked:', relpath)
                self.untracked += 1

            elif file.name == '.git':
                continue

            # check if dir is a submodule
            elif file.name in (
                    i[2] for i in
                    filter(lambda x: x[0] == '160000', tree_items_list)
            ):
                found_in_tree += 1

            # if dir isn't in tree_items_list
            elif file.name not in (
                    i[2] for i in
                    filter(lambda x: x[0] == '40000', tree_items_list)
            ):
                # untrack must be setted only if any children is not ignored
                result = self.handle_untracked_dir(file.path)
                if result:
                    print('Untracked:', relpath)
                    self.untracked += result

            # is_dir and is in tree
            else:
                found_in_tree += 1

                # filter entries with type == 40000
                trees = filter(lambda x: x[0] == '40000', tree_items_list)
                file_name_tuple = filter(lambda x: x[2] == file.name, trees)
                item_tuple = tuple(file_name_tuple)
                tree_hash = item_tuple[0][1]

                self.get_full_status(
                    file.path,
                    tree_hash
                )

        # endfor
        self.deleted += len(tree_items_list) - found_in_tree

    def handle_tracked_file(self, file, tree_items_list, relpath):
        local_found_in_tree = 0

        # get the item or it's staged
        for item in tree_items_list:
            if item[2] == file.name:
                break
        else:
            print('Staged:', relpath, "isn't under a commit.")
            self.staged += 1
            return 0  # or local_found_in_tree

        local_found_in_tree = 1

        mtime = self.index_tracked[relpath]
        if file.stat().st_mtime != mtime:
            print('Modified:', relpath)
            self.modified += 1
            return local_found_in_tree

        file_hash = self.get_hash_of_file(file.path)
        file_hash_in_tree = item[1]

        if file_hash_in_tree == file_hash:
            os.utime(file.path, (time.time(), mtime))
            return local_found_in_tree
        elif file_hash_in_tree == self.get_hash_of_file(file.path,
                                                        use_cr=True):
            os.utime(file.path, (time.time(), mtime))
            return local_found_in_tree

        print('Modified:', relpath)
        self.modified += 1
        return local_found_in_tree

    def handle_untracked_dir(self, dir_path: str) -> int:
        # although it is untracked, files here can be staged
        """
        Handles when `get_full_status` finds untracked directory.
        param `dir_name`: The path of the untracked directory.
        return: int: 1 if there was at least one not ignored file in directory.
                     0 if there was none.
        """
        try:
            directory = os.scandir(dir_path)
        except (PermissionError, NotADirectoryError):
            return 0

        sub_dir = []
        for file in directory:
            if file.name == '.git':
                return 0
            sub_dir.append(DirEntryWrapper(file))

        # if sub_file is in tracked, dont flag untracked dir
        staged_flag = False
        for sub_file in sub_dir:
            sub_relpath = sub_file.path.removeprefix(
                self.git_dir + '\\'
            ).replace('\\', '/', -1)

            if sub_relpath in self.index_tracked:
                print('Staged:', sub_relpath)
                self.staged += 1
                staged_flag = True

        if staged_flag:
            return 0

        local_untracked = 0
        for sub_file in sub_dir:
            if sub_file.is_dir():
                local_untracked += self.handle_untracked_dir(sub_file.path)

            if local_untracked:
                return 1

            if sub_file.is_file():
                if any(fnmatchcase(sub_file.name, pattern)
                       for pattern in self.ignored):
                    continue

                sub_relpath = sub_file.path.removeprefix(
                    self.git_dir + '\\'
                ).replace('\\', '/', -1)

                if any(fnmatchcase(
                        sub_relpath + '/'
                        if pattern.endswith('/') else sub_relpath,
                        pattern) for pattern in self.fixed):
                    continue
                return 1

        return 0

    def get_tree_items(self, last_cmmt: str) -> list | None:
        """
        Gets the last commit tree itens (type, hash and filename).
        param `last_cmmt`: The hash of the last commit.
        return: list | None: list of tuples with the type, hash and filenames,
                    None if tree is empty.
        """

        last_cmmt_obj = self.get_content_by_hash_loose(last_cmmt)
        if last_cmmt_obj is None:
            last_cmmt_obj = self.get_content_by_hash_packed(last_cmmt)

        if last_cmmt_obj is None:
            print("Couldn't get last commit object. Fallback.")
            raise FallbackError

        tree_hash = self.get_tree_hash_from_commit(last_cmmt_obj)
        print('Last commit tree hash:', tree_hash)

        from_pack = False
        tree_obj = self.get_content_by_hash_loose(tree_hash)
        if tree_obj is None:
            from_pack = True
            tree_obj = self.get_content_by_hash_packed(tree_hash)

        if tree_obj is None:
            print("Couldn't get last commit tree. Fallback.")
            raise FallbackError

        # there is a tree object but the git worktree is empty.
        if not tree_obj:
            return

        return self.parse_tree_object(tree_obj, from_pack)

    def get_tree_items_from_hash(self, tree_hash: str) -> list:
        """
        Gets the tree itens searching for its hash.
        param `tree_hash`: The hash of tree.
        return: list: list of tuples with the type, hash and filenames.
        """
        from_pack = False
        tree_obj = self.get_content_by_hash_loose(tree_hash)
        print('Searching loose', tree_hash)

        if tree_obj is None:
            from_pack = True
            tree_obj = self.get_content_by_hash_packed(tree_hash)
            print('Searching packed', tree_hash)

        if tree_obj is None:
            print('Not found. Fallback.')
            raise FallbackError

        tree_items_list = self.parse_tree_object(tree_obj, from_pack)
        print('Found:', *tree_items_list, sep='\n', end=n)

        return tree_items_list



import os
from collections import OrderedDict


def main(queries: str, width: int, height: int) -> str:
    if not queries:
        return 'Pass queries as arguments.'

    file = os.path.join(
        os.environ['APPDATA'],
        'Microsoft/Windows',
        'PowerShell/PSReadLine/ConsoleHost_history.txt'
    )
    with open(file, 'r', encoding='utf-8') as history:
        content = history.read()

    content = content.splitlines()
    content = OrderedDict.fromkeys(content[-2::-1])
    height -= 4
    height = max(height, 5)
    res = ''

    counter = 0
    printed = 0
    flag = False
    for item in content:

        # Checks the point when we're not printing anymore
        if not flag and counter >= height:
            printed = counter
            flag = True

        for query in queries.split():
            if query in item:

                # Check if isn't the first loop
                if counter and not flag:
                    res += '\n'

                counter += 1
                if len(item) > width:
                    counter += 1

                if not flag:
                    res += item.replace(query, f'\x1b[32m{query}\x1b[0m', -1)
                break

    if flag:
        res += ' ' * 10 + f'\x1b[33m[and {counter - printed} more]\x1b[0m'

    return res



import os

PORT = 5432
SEP = os.path.sep
APP_HOME = os.environ['LOCALAPPDATA'] + SEP + 'ShellServer'
CACHE_PATH = APP_HOME + SEP + 'ShellServerCache'
CONFIG_PATH = os.path.expanduser('~/.shellserver.toml')



from . import server

try:
    server.mainloop()
except Exception:
    import traceback
    from datetime import datetime
    from .__init__ import APP_HOME

    with open(APP_HOME + '/traceback', 'w') as out:
        print(datetime.now().ctime(), file=out)
        print(traceback.format_exc(), file=out)

    raise



from __future__ import annotations

"""
Low level operations on packfiles.
"""

import mmap
import zlib


# dict[str path, mmap.mmap]
MAPPED_CACHE = {}


def _get_buffer(path) -> mmap.mmap:
    if path in MAPPED_CACHE and not MAPPED_CACHE[path].closed:
        return MAPPED_CACHE[path]

    with open(path, 'rb') as raw:
        m = mmap.mmap(raw.fileno(), 0, access=mmap.ACCESS_READ)

    MAPPED_CACHE[path] = m
    return m


class Packs:
    def search_idx(
            self,
            idx_path: str,
            hash_: str,
            rt_offset=False
    ) -> bool | int | None:
        """
        Searches given `idx_path` for `hash_`.
        param `idx_path`: The path of the pack index file.
        param `rt_offset`: Should this function return the offset of given file
                           if it's found?
        return: bool | int | None:
            If `rt_offset` is set to False, returns boolean value
            representing the presence of hash in the index.
            If `rt_offset` is set to True, returns the integer
            representing the offset of hash in the pack file.
            If `rt_offset` is set to True, but hash wasn't found in
            the index file, returns None.
        """

        layer1_idx = int(hash_[:2], 16) - 1
        found = False

        file = _get_buffer(idx_path)

        if layer1_idx > 0:
            file.seek(8 + 4 * layer1_idx)
            files_before = int.from_bytes(file.read(4), 'big')
        else:
            files_before = 0

        file.seek(1028)
        total_files = int.from_bytes(file.read(4), 'big')

        file.seek(
            1032  # end of fanout_layer1
            + 20 * files_before  # each will have 20 bytes
        )

        while not found:
            file_hash_bytes = file.read(20)
            file_hash = hex(
                int.from_bytes(file_hash_bytes, 'big')
            )[2:].zfill(40)

            if hash_ == file_hash:
                found = True
                break

            elif int(file_hash[:2], 16) > layer1_idx + 1:
                break

            elif file.tell() >= 1032 + 20 * total_files:
                break

            files_before += 1

        if not rt_offset:
            return found

        if found:
            file.seek(
                1032
                + 20 * total_files  # jump layer2
                + 4 * total_files  # jump layer3
                + 4 * files_before
            )
            return int.from_bytes(file.read(4), 'big')

    def get_content_by_offset(
            self, pack_path: str, offset: int
    ) -> bytes | None:
        """
        Gets the content of object in given `pack_path` by its `offset`.
        param `pack_path`: The path to the pack file.
        param `offset`: The offset of the object.
        return: bytes | None: The content of the object,
                              None it is a delta object
        """

        file = _get_buffer(pack_path)
        file.seek(offset)

        int_ = int.from_bytes(file.read(1), 'big')
        binary = f'{int_:b}'.zfill(8)
        type_ = binary[1:4]
        obj_size = int_ & 0x0f
        bit_shift = 4

        if type_ == '110':  # delta type
            return

        msb = binary.startswith('1')
        # size = binary[4:]

        while msb:
            int_ = int.from_bytes(file.read(1), 'big')
            binary = f'{int_:b}'.zfill(8)
            obj_size |= (int_ & 0x7f) << bit_shift
            bit_shift += 7
            msb = binary.startswith('1')

        return zlib.decompress(file.read(obj_size + 11))

    def get_idx_of_pack(self, pack: str) -> str:
        """
        return pack.removesuffix('pack') + 'idx'
        """
        return pack.removesuffix('pack') + 'idx'



"""
This submodule provides an inline view of 'git status'
"""

import sys

from . import low, high

OBJ = high.High()


def gitstatus(
    target_path: str,
) -> tuple[str, str]:
    """
    Highest level function that gets an inline version of
    'git status'.
    param `target_path`: The path where the search for git info begins.
    return: tuple of two strings: The branch and the status of it.
    """

    if '--disable-git' in sys.argv:
        return None, None

    git_dir = low.get_dot_git(target_path)

    if git_dir is not None and low.exists_head(git_dir):
        branch = low.get_branch_on_head(git_dir)
    else:
        return None, None

    OBJ.init(git_dir, branch)

    try:
        if '--use-git' in sys.argv:
            raise high.FallbackError
        status = OBJ.status()

    except high.FallbackError:
        status = OBJ.parse_git_status()

    except Exception:
        if '--let-crash' in sys.argv:
            raise
        status = OBJ.parse_git_status()

    return branch, status



from __future__ import annotations

"""
Module with functions considered to be of low level,
that is, functions that doesn't call its siblings.
Same idea for the methods of Low class.
"""

import hashlib
import os
import subprocess
import zlib


def get_dot_git(target_path: str) -> str | None:
    """
    Searches backwards for .git
    param `target_path`: The absolute path from which the search begins.
    return: str | None: The directory of .git. None if it was not found.
    """

    git = os.path.join(target_path, '.git')
    if os.path.exists(git):
        return target_path

    parent = os.path.dirname(target_path)
    if parent == target_path:
        return

    return get_dot_git(parent)


def exists_head(git_dir: str) -> bool:
    """
    Checks if there are a 'HEAD' in git files.
    return: bool: Is there a 'HEAD' in .git?
    """

    head = os.path.join(git_dir, '.git/HEAD')
    return os.path.exists(head)


def get_branch_on_head(git_dir: str) -> str:
    """
    Parses 'HEAD' file to get the head branch.
    return: str: The name of the git branch.
    """

    head_path = os.path.join(git_dir, '.git/HEAD')

    with open(head_path) as head:
        content = head.read()

    return content[content.rindex('/') + 1:].strip()


class Low:
    def get_info_packs_content(self) -> list:  # [str]:
        """
        Checks the existence of a 'packs' file in git/objects
        and return its content.
        return: list: Content of packs by line.
        """

        packs = os.path.join(self.git_dir, '.git/objects/info/packs')
        if not os.path.exists(packs):
            return []

        with open(packs) as in_file:
            content = in_file.readlines()[::-1]

        counter = 0
        while counter < len(content):
            test = content[counter][2:].strip()
            if not test:
                del content[counter]
                continue

            content[counter] = test
            counter += 1

        return content

    def get_gitignore_content(self) -> list:  # [str]:
        """
        Checks the existence of a .gitignore in same level of .git
        and return its content
        return: list: Content of .gitignore by line.
        """

        gitignore = os.path.join(self.git_dir, '.gitignore')
        if not os.path.exists(gitignore):
            return []

        with open(gitignore) as in_file:
            content = in_file.readlines()

        counter = 0
        while counter < len(content):
            test = content[counter].strip()
            if not test:
                del content[counter]
                continue

            content[counter] = test
            counter += 1

        return content

    def get_last_commit_loose(self) -> str | None:
        """
        Gets the last commit's hash of a git repo.
        return: str | None: String of the last commit hash, None if there's no
                            last commit.
        """

        head_path = os.path.join(
            self.git_dir, f'.git/refs/heads/{self.branch}'
        )

        if not os.path.exists(head_path):
            return

        with open(head_path) as file:
            return file.read().strip()

    def get_info_refs_content(self) -> list[str]:
        """
        Checks the existence of a 'refs' file in .git/info
        and return its content
        return: list: Content of 'refs' by line.
        """

        refs = os.path.join(self.git_dir, '.git/info/refs')
        if not os.path.exists(refs):
            return []

        with open(refs) as in_file:
            content = in_file.readlines()

        for ind in range(len(content)):
            content[ind] = content[ind].strip()

        return content

    def get_exclude_content(self) -> list:  # [str]:
        """
        Checks the existence of a 'exclude' file inside .git/info
        and return its content.
        return: list: Content of exclude by line.
        """
        exclude = os.path.join(self.git_dir, '.git/info/exclude')
        if not os.path.exists(exclude):
            return []

        with open(exclude) as in_file:
            content = in_file.readlines()

        counter = 0
        while counter < len(content):
            test = content[counter].strip()
            if not test:
                del content[counter]
                continue

            content[counter] = test
            counter += 1

        return content

    def get_hash_of_file(self, file_path: bytes | str, use_cr=False) -> str:
        """
        Get the SHA1 hash in the same way Git would do.
        param `file_path`: str | bytes: The path to file.
        param `use_cr`: bool: Use CRLF for new line?
        """
        with open(file_path, 'rb') as in_file:
            content = in_file.read()

        if not use_cr:
            content = content.replace(b'\r\n', b'\n', -1)

        size = len(content)
        string = f"blob {size}\x00"
        hash_ = hashlib.sha1(string.encode() + content).hexdigest()

        return hash_

    def get_content_by_hash_loose(self, hash_: str) -> bytes | None:
        """
        Get the content of a loose file by its hash.
        param `hash`: The hash gotten from `get_hash_of_file` for files in repo
                      or in git files.
        return: bytes | None: The content of file, None if it was not found.
        """

        path = os.path.join(
            self.git_dir, f'.git/objects/{hash_[:2]}/{hash_[2:]}'
        )
        if not os.path.exists(path):
            return

        with open(path, 'rb') as in_file:
            content = in_file.read()

        # cannot decode here
        # in blob we can take bytes or str
        # in tree, only bytes
        return zlib.decompress(content)

    def get_tree_hash_from_commit(self, cmmt_obj: bytes) -> str:
        """
        Get the tree hash from an commit object.
        param `cmmt_obj`: bytes: A git commit object.
        return: str: The tree hash.
        """

        first_line = cmmt_obj.splitlines()[0].decode()
        tree_hash = first_line[first_line.index('tree') + 5:].strip()

        return tree_hash

    def get_status_string(
            self, status: tuple[int, int, int, int]
    ) -> str | None:
        """
        Get the string version of given Git status
        param `status`: tuple: untracked, staged, modified and deleted sums.
        return: str | None: String of status, None if not any value is above 0.
        """
        if not any(status):
            return

        symbols = ('?', '+', 'm', 'x')

        return ' '.join(
            symb + str(stat)
            for symb, stat in zip(symbols, status)
            if stat
        )

    def parse_git_status(self) -> str:
        """
        Parses `git status -s` returning tuple of untracked, staged,
        modified and deleted sums.
        return: str: String of status.
        """
        data, err = subprocess.Popen(
            f'git -C {self.git_dir} status -s --porcelain',
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            creationflags=subprocess.CREATE_NO_WINDOW
        ).communicate()

        data = data.decode().splitlines()
        data = [line[:2].strip() for line in data]

        possibles = list(set(data))

        res = [mark + str(data.count(mark)) for mark in possibles]

        return ' '.join(res)

    def parse_tree_object(
            self, data: bytes, from_pack: bool = False
    ) -> list:  # [tuple[str, str, str]] | None:
        """
        Parses the content of a git tree object and returns a list
        of its values
        param `data`: The content decompressed of tree object
        return: list | None: list of tuples containing the type, hash and
                the filename in tree, None if tree is empty.
        """

        if not from_pack:
            if not data[data.index(b'\x00') + 1:]:
                return

            # first remove everything before first null byte
            start = data.index(b'\x00') + 1
        else:
            start = 0

        res = []

        while True:
            # cant split(b'\x00), hash might have it
            # cant split(' '), filename might have it
            stop = data.index(b'\x00', start)

            piece = data[start: stop]
            sep = piece.index(b' ')

            type_ = piece[:sep]
            filename = piece[sep + 1:]

            start = stop + 21  # next start

            hexa = data[stop + 1:start]
            hexa = f'{int.from_bytes(hexa, "big"):x}'.zfill(40)

            res.append((type_.decode(), hexa, filename.decode().lower()))

            if not data[start + 1:]:
                break

        return res



import json
import os
import winreg


def _get_actual_value(subkey, name):

    read_handle_hkey = winreg.OpenKeyEx(
        winreg.HKEY_CURRENT_USER, subkey, 0, winreg.KEY_READ
    )

    actual_value = winreg.QueryValueEx(
        read_handle_hkey, name
    )[0]

    winreg.CloseKey(read_handle_hkey)

    return actual_value


def night_light_change():
    """
    Toggles blue light reduction in Windows.
    """

    # This function is a "translation" from:
    # https://github.com/inamozov/DisplayTest/blob/master/DisplayTest.cpp

    SUB_KEY = (
        'Software\\Microsoft\\Windows\\CurrentVersion\\CloudStore\\'
        'Store\\DefaultAccount\\Current\\'
        'default$windows.data.bluelightreduction.bluelightreductionstate\\'
        'windows.data.bluelightreduction.bluelightreductionstate'
    )
    NAME = 'Data'

    value = bytearray(_get_actual_value(SUB_KEY, NAME))

    on = value[18] == 21
    size = 41 if on else 43
    if on:
        for i in range(10, 15):
            if value[i] != 255:
                value[i] += 1
                break
        value[18] = 19
        for i in range(24, 22, -1):
            for j in range(i, size - 2):
                value[j] = value[j + 1]
    else:
        for i in range(10, 15):
            if value[i] != 255:
                value[i] += 1
                break
        value[18] = 21
        n = 0
        while n < 2:
            for i in range(size - 1, 23, -1):
                value[i] = value[i - 1]
            n += 1
        value[23] = 16
        value[24] = 0

    set_handle_hkey = winreg.OpenKeyEx(
        winreg.HKEY_CURRENT_USER, SUB_KEY, 0, winreg.KEY_SET_VALUE
    )

    winreg.SetValueEx(
        set_handle_hkey, NAME, 0, winreg.REG_BINARY, value
    )
    winreg.CloseKey(set_handle_hkey)


def system_change():
    SUB_KEY = ('SOFTWARE\\Microsoft\\Windows\\'
               'CurrentVersion\\Themes\\Personalize')
    NAME = "SystemUsesLightTheme"

    actual_value = _get_actual_value(SUB_KEY, NAME)

    target = 0 if actual_value == 1 else 1

    set_handle_hkey = winreg.OpenKeyEx(
        winreg.HKEY_CURRENT_USER, SUB_KEY, 0, winreg.KEY_SET_VALUE
    )

    winreg.SetValueEx(
        set_handle_hkey, NAME, 0, winreg.REG_DWORD, target
    )
    winreg.SetValueEx(
        set_handle_hkey, "AppsUseLightTheme", 0, winreg.REG_DWORD, target
    )

    winreg.CloseKey(set_handle_hkey)


def windows_terminal_change(config):
    DEFAULT = os.path.join(
        os.environ['LOCALAPPDATA'],
        'Packages\\Microsoft.WindowsTerminal_8wekyb3d8bbwe',
        'LocalState\\settings.json'
    )
    PREVIEW = os.path.join(
        os.environ['LOCALAPPDATA'],
        'Packages\\Microsoft.WindowsTerminalPreview_8wekyb3d8bbwe',
        'LocalState\\settings.json'
    )
    UNPACKAGED = os.path.join(
        os.environ['LOCALAPPDATA'],
        'Microsoft\\Windows Terminal\\settings.json'
    )

    for const in (DEFAULT, PREVIEW, UNPACKAGED):
        if os.path.exists(const):
            path = const
            break
    else:
        return

    with open(path, 'r') as orig_file:
        file = json.load(orig_file)

    # settings should aways have profiles and defaults
    defaults = file.get('profiles').get('defaults')

    actual_scheme = defaults.get('colorScheme')
    if actual_scheme is None:
        SUB_KEY = ('SOFTWARE\\Microsoft\\Windows\\'
                   'CurrentVersion\\Themes\\Personalize')
        NAME = "SystemUsesLightTheme"

        if _get_actual_value(SUB_KEY, NAME):  # is light?
            defaults['colorScheme'] = config.light_theme
        else:
            defaults['colorScheme'] = config.dark_theme

    elif actual_scheme == config.light_theme:
        defaults['colorScheme'] = config.dark_theme
    else:
        defaults['colorScheme'] = config.light_theme

    with open(path, 'w') as mod_file:
        json.dump(file, mod_file, indent=4)


def all():
    system_change()
    windows_terminal_change()
    night_light_change()



import socket

from .__init__ import APP_HOME, PORT, SEP, CONFIG_PATH

# Quit program as soon as possible
try:
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.bind(('localhost', PORT))
except OSError:
    raise SystemExit


import gc
import os
import subprocess
import sys
import threading
import time
from dataclasses import dataclass

try:
    import tomllib
except ImportError:
    import tomlkit as tomllib

from win_basic_tools import Ls

from . import histdb, theme, gitstatus
from .classes import DirCache, Dispatcher
from .style import style


gc.disable()


@dataclass
class Config:
    dark_theme: str = 'Tango Dark'
    light_theme: str = 'Solarized Light'


def shell_manager(entry: str, addr) -> None:
    global clients

    if entry.startswith('Init'):
        shell = entry.removeprefix('Init')
        dispatcher.register(addr, shell)
        clients += 1
        completions = ';'.join(item[2] for item in cache.dirs)
        dispatcher.send_through(sock, completions, addr)

    elif entry.startswith('Set'):
        option = '--' + entry.removeprefix('Set')
        if option == '--enable-git':
            try:
                sys.argv.remove('--disable-git')
            except ValueError:
                pass
            return
        if option == '--verbose':
            sys.stdout = sys.__stdout__
        sys.argv.append(option)

    elif entry == 'Exit':
        clients -= 1
        if clients <= 0:
            cache.finish()
            raise SystemExit

    elif entry == 'Kill':
        cache.finish()
        raise SystemExit


def zsearch(entry: str, addr: int) -> None:
    # ./path/ -> path
    # .path -> .path
    entry = entry.strip(SEP).removeprefix('.' + SEP).lower()
    result = cache.get(entry)
    dispatcher.send_through(sock, result, addr, prepare=False)


def fzfsearch(entry: str, addr: int) -> None:  # will receive two calls by 'pz'
    if not entry:  # if its the first
        result = '\n'.join([i[1] for i in cache.dirs])
        dispatcher.send_through(sock, result, addr)
    else:
        cache.update_by_full_path(entry)
        cache.sort()


def list_directory(entry: str, addr: int) -> None:
    opt, path = entry.split(';')
    path = path.removeprefix('Microsoft.PowerShell.Core\\FileSystem::')

    # Ls class can handle symlink
    ls = Ls(opt, path, to_cache=True)
    ls.echo()

    dispatcher.send_through(sock, ls.out.data, addr)


def theme_manager(entry: str, addr) -> None:
    try:
        if entry in ('all', ''):
            theme.all()
        elif entry == 'terminal':
            theme.windows_terminal_change(config)
        elif entry == 'blue':
            theme.night_light_change()
        elif entry == 'system':
            theme.system_change()
    except Exception:
        pass


def history_search(entry: str, addr) -> None:
    queries, width, height = entry.split(';')
    width, height = int(width), int(height)

    res = histdb.main(queries, width, height)
    dispatcher.send_through(sock, res, addr)


def _get_paths(curdir):
    # Microsoft.PowerShell.Core\Registry::
    curdir = curdir.removeprefix('Microsoft.PowerShell.Core\\')
    # os.realpath can handle \\wsl$\Distro
    curdir = curdir.removeprefix('FileSystem::')

    # Python 3.10+
    # try:
    #     link = os.path.realpath(curdir, strict=True)
    # except OSError:
    #     #  Env:, Function:, Alias:, ...
    #     link = curdir

    # Workaround. Can lead to errors
    link = os.path.realpath(curdir)
    if not os.path.exists(link):
        link = curdir

    target_dir = curdir if curdir == link else link

    if home in curdir:
        curdir = '~' + curdir.removeprefix(home)
    if home in link:
        link = '~' + link.removeprefix(home)

    return curdir, link, target_dir


def scan(entry, addr):
    no_error = int(entry[0])
    curdir, width, duration = entry[1:].split(';')

    width = int(width)
    duration = round(float(duration.replace(',', '.')), 1)

    # must add before cleanup because Set-Location needs full 'path'
    if curdir != home:
        cache.add(curdir)

    curdir, link, target_dir = _get_paths(curdir)

    brackets = []
    after_brackets = []

    branch, status = gitstatus.gitstatus(target_dir)
    if branch is not None:
        brackets.append(f'Branch;{branch}')
        if status:
            brackets.append(f'Status;{status}')

    git_flag = bool(branch)

    try:
        directory = os.scandir(target_dir)
    except PermissionError:
        brackets.append('Error;Permission')
        directory = tuple()
    except OSError:
        directory = tuple()

    for file in directory:
        if (py_not not in brackets
           and file.name.endswith(('.py', '.pyc', '.pyw', '.pyd'))):
            brackets.append(py_not)

        for exe in exes_with_version:
            notation = f'{map_lang_name[exe]};{exes_with_version[exe]}'
            condition1 = notation not in brackets
            if condition1 and file.name.endswith(map_suffix[exe]):
                brackets.append(notation)

        for exe in exes_to_search:  # those who don't have version
            notation = map_lang_name[exe]
            if (notation not in after_brackets
               and file.name.endswith(map_suffix[exe])):
                after_brackets.append(notation)

    prompt = style(
        curdir, link, git_flag, brackets,
        after_brackets, no_error, width, duration
    )

    dispatcher.send_through(sock, prompt, addr)


def mainloop():
    while 1:
        entry, addr = sock.recvfrom(4096)
        init = time.perf_counter()
        entry = entry.decode()
        prot = entry[0]

        functionalities[prot](entry[1:], addr)

        if '--verbose' in sys.argv:
            took = round(time.perf_counter() - init, 5)
            print(f'Took: {took}s')

        for mmap in gitstatus.packs.MAPPED_CACHE.values():
            mmap.close()

        gc.collect()


#
# Updating the executables that will be searched starts here
#


def get_version(exe) -> None:
    try:
        out, err = subprocess.Popen(
            f'{exe} --version',
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            creationflags=subprocess.CREATE_NO_WINDOW
        ).communicate()
    except FileNotFoundError:
        return

    if err:
        return

    if exe in ('g++', 'gcc'):
        out = out.splitlines()[0]
        out = out[out.rindex(b')') + 2:]
        exes_with_version.update({exe: out.strip().decode()})
        exes_to_search.remove(exe)
    elif exe == 'node':
        exes_with_version.update({exe: out.strip().decode()[1:]})
        exes_to_search.remove(exe)
    elif exe == 'pwsh':
        exes_with_version.update({exe: out.split()[-1].strip().decode()})
        exes_to_search.remove(exe)


exes_to_search = ['node', 'g++', 'gcc', 'lua', 'pwsh']

map_suffix = {
    'node': '.js',
    'g++': '.cpp',
    'gcc': '.c',
    'lua': '.lua',
    'pwsh': ('.ps1', '.psd1', '.psm1')
}
map_lang_name = {
    'node': 'Node',
    'g++': 'Cpp',
    'gcc': 'C',
    'lua': 'Lua',
    'pwsh': 'Pwsh'
}

#
# Stop here
#

os.makedirs(APP_HOME, exist_ok=True)

entries = map(str, range(1, 8))
defined_functions = (
    scan,
    shell_manager,
    zsearch,
    fzfsearch,
    list_directory,
    theme_manager,
    history_search
)
functionalities = dict(zip(entries, defined_functions))

clients = 0
cache = DirCache()
dispatcher = Dispatcher()
home = os.path.expanduser('~')

py_version = sys.version
py_version = py_version[:py_version.index(' ')]
py_not = 'Python' + ';' + py_version  # notation

exes_with_version = {}

for exe in exes_to_search:
    threading.Thread(target=get_version, args=(exe,)).start()

if os.path.exists(CONFIG_PATH):
    with open(CONFIG_PATH, 'rb') as file:
        toml = tomllib.load(file)
else:
    toml = {}

config = Config(**toml)



import os
import pickle
import threading

from .__init__ import CACHE_PATH, SEP


class DirCache:
    __slots__ = 'dirs', 'flag'

    # dirs = list[list[precedence: int, abs_path: str, short_path: str]]
    def __init__(self):
        if os.path.exists(CACHE_PATH):
            with open(CACHE_PATH, 'rb') as in_file:
                self.dirs = pickle.load(in_file)
            self._clear()
        else:
            self.dirs = []

            # Need to create file or cache will never save
            # if it doesn't exist.
            # Need to put something there or server might break
            # when pickle tries to read an empty file.
            with open(CACHE_PATH, 'wb') as in_file:
                pickle.dump([], in_file)

        self.flag = False  # New paths were added to cache?

    def add(self, path: str) -> None:
        for item in self.dirs:
            if item[1] == path:
                break
        else:
            out = path[path.rindex(SEP) + 1:].lower()
            if not out:
                out = path.strip(SEP).lower()

            self.dirs.append([0, path, out])

            # just want to set flag if it is false to start thread
            # that will capture WM_SAVE_YOURSELF
            if not self.flag:
                self.flag = True
                threading.Thread(target=self._signal_cap, daemon=True).start()

    def get(self, rel_path: str) -> str:
        for item in self.dirs:
            if item[2] == rel_path:
                return item[1]
        return ''

    def update_by_full_path(self, full_path: str) -> None:
        for ind in range(len(self.dirs)):
            if self.dirs[ind][1] == full_path:
                self.dirs[ind][0] += 1
                break

    def finish(self):
        self._clear()
        self._save()

    def sort(self) -> None:
        self.dirs.sort(key=lambda x: x[0], reverse=True)

    def _save(self) -> None:
        # if there were calls to clear the cache during this runtime
        if not os.path.exists(CACHE_PATH):
            return
        if self.flag:
            with open(CACHE_PATH, 'wb') as out:
                pickle.dump(self.dirs, out, protocol=5)

    def _clear(self) -> None:
        aux = []
        for item in self.dirs:
            if not os.path.exists(item[1]):
                aux.append(item)
        if aux:
            self.flag = True
        for item in aux:
            self.dirs.remove(item)

    def _signal_cap(self):
        import tkinter as tk

        def finish():
            self.finish()
            self.flag = False  # will make new threads possible
            raise SystemExit

        root = tk.Tk()
        root.withdraw()
        root.protocol('WM_SAVE_YOURSELF', finish)
        root.mainloop()


class Dispatcher:
    """
    This class will hold the specifics convertions
    necessary to each shell, apply them and send
    data through socket.
    """
    __slots__ = 'addrs', 'funcs'

    def __init__(self):
        self.addrs = {}
        self.funcs = {'pwsh': self._pwsh_func}

    def register(self, addr: int, shell: str):
        self.addrs.update({addr: shell})

    def send_through(self, sock, data: str,
                     addr: tuple[str, int], *, prepare=True):

        shell = self.addrs.get(addr, 'pwsh')
        func = self.funcs.get(shell)

        if prepare and func is not None:
            data = func(data)

        while data[60_000:]:
            msg = '1' + data[:60_000]
            sock.sendto(msg.encode(), addr)
            data = data[60_000:]

        msg = '0' + data

        sock.sendto(msg.encode(), addr)

    def _pwsh_func(self, data: str):
        return data.replace('$', '`$', -1)



from __future__ import annotations

"""
Module with functions considered of medium level of abstraction
(functions here calls low functions and/or its siblings).
"""

import sys

from . import low
from . import packs

os = low.os


class IndexTooBigError(Exception):
    pass


class Medium(low.Low, packs.Packs):
    def set_packs(self) -> None:
        """
        Sets the `packs_list` attribute.
        return: None.
        """
        # info/packs is not needed
        pack_def_path = '.git/objects/pack'

        packs_list = self.get_info_packs_content()
        if packs_list:
            self.packs_list = [
                os.path.join(
                    self.git_dir, pack_def_path, i
                ) for i in packs_list
            ]
            return

        packs = os.path.join(self.git_dir, pack_def_path)
        if not os.path.exists(packs):
            self.packs_list = []
            return

        packs_list = os.scandir(packs)
        self.packs_list = [
            i.path for i in packs_list if i.name.endswith('.pack')
        ]

    def set_index_tracked(self) -> None:
        """
        Sets `index_tracked` attribute.
        Modified and simpler version of gin.
        https://github.com/sbp/gin
        return: None.
        """
        index_path = os.path.join(self.git_dir, '.git/index')
        if not os.path.exists(index_path):
            self.index_tracked = tuple()
            return

        with open(index_path, 'rb') as f:

            def readStrUntil(delim):
                ret = []
                while True:
                    b = f.read(1)
                    if b == '' or b == delim:
                        return b"".join(ret).decode()

                    ret.append(b)

            constant = f.read(4)
            version = int.from_bytes(f.read(4), 'big')
            if constant != b'DIRC' or version not in (2, 3):
                self.index_tracked = None

            entries = int.from_bytes(f.read(4), 'big')
            res = {}

            for entry in range(entries):
                f.read(8)
                mtime = int.from_bytes(f.read(4), 'big')
                mtime += int.from_bytes(f.read(4), 'big') / 1000000000
                f.read(44)

                flags = int.from_bytes(f.read(2), 'big')
                namelen = flags & 0xfff
                extended = flags & (0b0100_0000 << 8)

                entrylen = 62

                if extended:
                    f.read(2)
                    entrylen += 2

                if namelen < 0xfff:
                    name = f.read(namelen).decode()
                    entrylen += namelen
                else:
                    name = readStrUntil('\x00')
                    entrylen += 1

                res[name.lower()] = mtime

                give_up = '--wait' not in sys.argv
                if give_up and len(res) > 2500:
                    raise IndexTooBigError

                padlen = (8 - (entrylen % 8)) or 8
                f.read(padlen)

        self.index_tracked = res

    def set_split_ignored(self) -> None:
        """
        Sets `fixed` and `ignored` attributes.
        `fixed` represents the patterns to ignore as full paths.
        `ignored` represents the patterns to ignore as relative paths.
        return: None.
        """
        raw_ignored = self.get_exclude_content()
        raw_ignored += self.get_gitignore_content()
        raw_ignored = [
            i.strip().lower()
            for i in raw_ignored
            if not i.strip().startswith('#')
        ]

        ignored = [
            i for i in raw_ignored
            if '/' not in i
        ]

        fixed = [i for i in raw_ignored if i not in ignored]

        self.fixed, self.ignored = fixed, ignored

    def get_content_by_hash_packed(self, hash_: str) -> bytes | None:
        """
        Gets the content of an object by its hash in any packfiles present.
        param `hash_`: The hash to be searched in packfiles.
        return: bytes | None: The content of object, None if it was not Found.
        """
        content = None

        for pack in self.packs_list:
            idx = self.get_idx_of_pack(pack)
            offset = self.search_idx(idx, hash_, rt_offset=True)
            if offset is None:
                continue
            content = self.get_content_by_offset(pack, offset)
            if content:
                break

        return content

    def get_last_commit_packed(self) -> str | None:
        """
        Gets the last commit's hash of a git repo by a packed perspective.
        return: str | None: String of the last commit hash, None if there's no
                            last commit packed.
        """
        info_refs = self.get_info_refs_content()
        last_commt_hash = None

        for line in info_refs:
            a, b = line.strip().split()
            if b == f'refs/heads/{self.branch}':
                last_commt_hash = a
                break

        return last_commt_hash

    def get_last_commit_hash(self) -> str | None:
        """
        Gets the last commit's hash of a git repo.
        return: str | None: String of the last commit hash, None if there's no
                            last commit.
        """
        last_cmmt = self.get_last_commit_loose()

        if last_cmmt is None:
            last_cmmt = self.get_last_commit_packed()

        return last_cmmt



import os
import threading
from datetime import datetime

import darkdetect

from .__init__ import SEP


def darkdetect_callback(arg):
    global colors_map

    if colors_map is dark_colors:
        colors_map = light_colors
    else:
        colors_map = dark_colors


def hide_first_dirs(path: str, from_depth: int) -> str:
    if path.count(SEP) <= from_depth:
        return path

    first_slice_end = path.index(SEP) + 1

    second_slice_start = len(path) - 1
    for _ in range(from_depth):
        second_slice_start = path.rindex(SEP, 0, second_slice_start)

    removed = path[first_slice_end:second_slice_start]
    ast = '*' * (1 + int(removed.count(SEP) > 0))

    res = path[:first_slice_end] + ast + path[second_slice_start:]
    return res


def reduce_big_path_names(path: str, max_size: int) -> str:
    names = path.split(SEP)

    for idx in range(len(names)):
        size = len(names[idx])
        if size > max_size:
            middle = int(max_size // 2)
            end_slice = middle - 1
            start_slice = size - middle + 1

            names[idx] = (
                names[idx][:end_slice]
                + '...'
                + names[idx][start_slice:]
            )

    return SEP.join(names)


def resolve_duration(seconds: float) -> str:
    # return like: 2h, 1h3m, 4m, 2m5s, 58.4s
    minutes = int(seconds / 60)
    hours, minutes = int(minutes / 60), minutes % 60
    seconds = seconds % 60

    res = ''
    for val, unit in zip((hours, minutes, seconds), ('h', 'm', 's')):
        if unit == 's' and not res and val > 2:
            res += f'{val}{unit}'
        elif unit == 's' and 'h' in res:
            break
        elif val and ((unit == 's' and (val > 2 or res)) or unit != 's'):
            res += f'{int(val)}{unit}'

    return res


def resolve_clocks(duration: float, width: int) -> str:
    res = resolve_duration(duration)
    if res:
        res = symbols_map['Duration'] + ' ' + res

    if width - len(res) - 11 >= 0:
        time_str = datetime.now().strftime('%H:%M:%S')
        if res:
            res += ' ' + symbols_map['Clock'] + ' ' + time_str
        else:
            res += symbols_map['Clock'] + ' ' + time_str

    return res


def get_cwd_and_link(cwd, link):
    equal = cwd == link

    if cwd == '~':
        cwd = ' ' + cwd
    elif cwd.startswith(os.environ['HOMEDRIVE']):
        cwd = ' ' + cwd
    else:
        cwd = ' ' + cwd

    cwd = hide_first_dirs(cwd, 5)
    cwd = reduce_big_path_names(cwd, 25)

    result = colors_map['Cwd'] + cwd + colors_map['Reset'] + ' '

    if not equal:
        link = hide_first_dirs(link, 5)
        link = reduce_big_path_names(link, 25)
        result += (
            colors_map['Link']
            + ' '
            + colors_map['Cwd']
            + link
            + colors_map['Reset']
            + ' '
        )

    return result


def get_brackets(brackets, after):
    result = ''
    for item in brackets:
        symbol, text = item.split(';')

        result += '[' + colors_map[symbol] + symbols_map[symbol]
        if symbol != 'Status':
            result += ' '
        result += text + colors_map['Reset'] + '] '

    for item in after:
        result += (
            colors_map[item] + symbols_map[item] + colors_map['Reset'] + ' '
        )

    # remove last empty space, new line might be in there
    result = result.strip()
    return result


def style(
    cwd: str,
    link: str,
    git: int,  # | bool,
    brackets: list,
    after: list,
    no_error: int,  # | bool,
    width: int,
    duration: float
) -> str:

    result_cwd = get_cwd_and_link(cwd, link)

    result_git = ''
    if git:
        result_git += (
            colors_map['Git'] + symbols_map['Git'] + colors_map['Reset'] + ' '
        )

    result_brackets = get_brackets(brackets, after)

    # get the lenght of written so far
    raw = result_cwd + result_git + result_brackets
    for item in colors_map:
        raw = raw.replace(colors_map[item], '', -1)

    width -= (
        len(raw)
        + int('🌙' in raw)  # because it's len == 1, but uses two spaces
        + 1  # new line at the end
    )

    result_clocks = resolve_clocks(duration, width)
    if result_clocks:
        result_brackets += ' '
        width -= 1

    width -= len(result_clocks) - int('🕓' not in result_clocks)
    result = (
        result_cwd + result_git + result_brackets
        + ' ' * width + result_clocks
    )

    if width > 0:
        result += '\n'

    if no_error:
        result += colors_map['No_error']
    else:
        result += colors_map['Error']

    result += '❯ ' + colors_map['Reset']

    return result


symbols_map = {
    'Git': '',
    'Branch': '',
    'Python': '',
    'Lua': '🌙',
    'Node': '',
    'C': '',
    'Cpp': '',
    'Pwsh': '',  # maybe  
    'Status': '',
    'Clock': '🕓',
    'Duration': '',
    'Error': '✘'
}

light_colors = {
    'Cwd': "\x1b[36m",
    'Link': "\x1b[90m",
    'Git': "\x1b[31m",
    'Branch': "\x1b[35m",
    'Python': "\x1b[33m",
    'Lua': "\x1b[34m",
    'Node': "\x1b[32m",
    'C': "\x1b[34m",
    'Cpp': "\x1b[34m",
    'Pwsh': "\x1b[34m",
    'Status': "\x1b[31m",
    'Reset': "\x1b[0m",
    'No_error': "\x1b[32m",
    'Error': "\x1b[31m"
}

dark_colors = {
    'Cwd': "\x1b[96m",
    'Link': "\x1b[90m",
    'Git': "\x1b[91m",
    'Branch': "\x1b[95m",
    'Python': "\x1b[93m",
    'Lua': "\x1b[94m",
    'Node': "\x1b[32m",
    'C': "\x1b[34m",
    'Cpp': "\x1b[94m",
    'Pwsh': "\x1b[94m",
    'Status': "\x1b[91m",
    'Reset': "\x1b[0m",
    'No_error': "\x1b[92m",
    'Error': "\x1b[91m"
}

colors_map = dark_colors if darkdetect.isDark() else light_colors

threading.Thread(
    target=darkdetect.listener, args=(darkdetect_callback,), daemon=True
).start()
